#include "lexer_states.h"
#include "aliases.h"
#include <ctype.h>

// clang-format off
typedef enum {
	// GENERAL STATES
  LEXER_STATE_START,
  LEXER_STATE_SUCCESS,
  LEXER_STATE_ERROR,
  LEXER_STATE_VALUE,
  LEXER_STATE_KEYWORD,
	// COLLECTION STATES
	LEXER_STATE_OBJECT,
	LEXER_STATE_ARRAY,
	// OBJECT STATES
	LEXER_STATE_KEY_START,
	LEXER_STATE_KEY_END,
	// NUMBER STATES
	LEXER_STATE_NUMBER_START,
  LEXER_STATE_DECIMAL,
  LEXER_STATE_NUMBER,
  LEXER_STATE_FRACTION,
  LEXER_STATE_EXPONENT,
  LEXER_STATE_EXP_SIGN,
  LEXER_STATE_POWER,
  LEXER_STATE_NUMBER_END,
	// STRING STATES
	LEXER_STATE_STRING,
	LEXER_STATE_ESCAPE_SEQUENCE,
	LEXER_STATE_UNICODE_HEX,

  COUNT_LEXER_STATES,
} lexer_state_t;
// clang-format on

struct lexer {
  lexer_state_t current;
  u64 line;
  u64 column;
};

void lexer_state_machine(lexer_t *lexer, char input);
lexer_state_t handle_lexer_start(char input);
lexer_state_t handle_object(char input);
lexer_state_t handle_character(char input);
lexer_state_t handle_escape_sequence(char input);

void lexer_state_machine(lexer_t *lexer, char input) {
  switch (lexer->current) {
  case LEXER_STATE_START:
    lexer->current = handle_lexer_start(input);
    break;
  case LEXER_STATE_SUCCESS:
    break;
  case LEXER_STATE_ERROR:
    break;
  case LEXER_STATE_VALUE:
    break;
  case LEXER_STATE_KEYWORD:
    break;
  case LEXER_STATE_OBJECT:
    lexer->current = handle_object(input);
    break;
  case LEXER_STATE_ARRAY:
    lexer->current = LEXER_STATE_VALUE;
    break;
  case LEXER_STATE_KEY_START:
    lexer->current = LEXER_STATE_STRING;
    break;
  case LEXER_STATE_KEY_END:
    break;
  case LEXER_STATE_NUMBER_START:
    break;
  case LEXER_STATE_DECIMAL:
    break;
  case LEXER_STATE_NUMBER:
    break;
  case LEXER_STATE_FRACTION:
    break;
  case LEXER_STATE_EXPONENT:
    break;
  case LEXER_STATE_EXP_SIGN:
    break;
  case LEXER_STATE_POWER:
    break;
  case LEXER_STATE_NUMBER_END:
    break;
  case LEXER_STATE_STRING:
    lexer->current = handle_character(input);
    break;
  case LEXER_STATE_ESCAPE_SEQUENCE:
    lexer->current = handle_escape_sequence(input);
    break;
  case LEXER_STATE_UNICODE_HEX:
    break;
  default:
    lexer->current = LEXER_STATE_ERROR;
  }
}

lexer_state_t handle_lexer_start(char input) {
  if (isspace(input)) {
    return LEXER_STATE_START;
  }

  switch (input) {
  case '{':
    return LEXER_STATE_OBJECT;
  case '[':
    return LEXER_STATE_ARRAY;
  }

  return LEXER_STATE_ERROR;
}

lexer_state_t handle_object(char input) {
  if (isspace(input)) {
    return LEXER_STATE_OBJECT;
  } else if (input == '"') {
    return LEXER_STATE_KEY_START;
  }

  return LEXER_STATE_ERROR;
}

lexer_state_t handle_character(char input) {
  switch (input) {
  case '\\':
    return LEXER_STATE_ESCAPE_SEQUENCE;
  case '"':
    return LEXER_STATE_STRING; // WHAT TO RETURN HERE?
  }

  return LEXER_STATE_STRING;
}

lexer_state_t handle_escape_sequence(char input) {
  switch (input) {
  case '"':
  case '/':
  case '\\':
  case 'b':
  case 'f':
  case 'n':
  case 'r':
  case 't':
    return LEXER_STATE_STRING;
  case 'u':
    return LEXER_STATE_UNICODE_HEX;
  }

  return LEXER_STATE_ERROR;
}
